from .CIFRank import CIFRank
from .gFair import gFair
from .iFair import iFair
from .LFR import LFR

fairness_method_mapping = {
    'CIFRank': CIFRank,
    'gFair': gFair,
    'iFair': iFair,
    'LFR': LFR
}